Module:       postscript-duim
Synopsis:     DUIM postscript backend
Author:       Scott McKay, Andy Armstrong
Copyright:    Original Code is Copyright (c) 1995-2004 Functional Objects, Inc.
              All rights reserved.
License:      Functional Objects Library Public License Version 1.0
Dual-license: GNU Lesser General Public License
Warranty:     Distributed WITHOUT WARRANTY OF ANY KIND

/// Font Metrics for Apple Laser Writer builtin fonts

// This actually reads metrics for the Apple LaserWriter.
// It initializes '*char-width-tables*' for a font in a "normalized" size.
// The real size gets computed as we need it by 'initialize-fcs'.
//--- Some day we should make the font width table stuff device independent.
define method setup-laserwriter-metrics
    (name, family, weight, slant, scale, size-kludge, box, font-info)
  let scale = as(<single-float>, scale);
  let width-table = #f;
  if (size-kludge)
    width-table := size-kludge
  else
    width-table := make(<simple-vector>, size: 256);
    for (char-info in font-info)
      let char-code = first(char-info);
      let char-width = second(char-info) / scale;
      width-table[char-code] := char-width
    end
  end;
  let key = list(family, weight, slant);
  // Check family to make sure we found the name
  let old = find-pair(*char-width-tables*, key, test: \=);
  let ascent  = abs(box[3] / scale);
  let descent = abs(box[1] / scale);
  let height  = ascent + descent;
  let data = list(name, width-table, height, ascent, descent);
  if (old)
    tail(old) := data
  else
    add!(*char-width-tables*, pair(key, data))
  end
end method setup-laserwriter-metrics;


// Note that the X values in the bounding box are incorrect.  I don't
// know how this data was generated (it was inherited from the Genera
// LGP2 support) but it is apparent that the font box information
// (/FontBBox element of a PostScript fnt dictionary) was not scaled by
// the font's matrix (/FontMatrix).  For the fonts that I tested (using
// the PostScript program at the end of this file), the Y values were
// correct but some of the X values were not.  Fortunately nothing uses
// them.  [1/9/90 naha]

setup-laserwriter-metrics
  ("Times-Roman", #"serif", #"normal", #"roman",
   1000, #f, #[-167, -252, 1004, 904],
   #[#[#o040, 250.0, "space"],
     #[#o041, 333.0, "exclam"],
     #[#o042, 408.0, "quotedbl"],
     #[#o043, 500.0, "numbersign"],
     #[#o044, 500.0, "dollar"],
     #[#o045, 833.0, "percent"],
     #[#o046, 778.0, "ampersand"],
     #[#o047, 333.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 500.0, "asterisk"],
     #[#o053, 564.0, "plus"],
     #[#o054, 250.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 250.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 500.0, "zero"],
     #[#o061, 500.0, "one"],
     #[#o062, 500.0, "two"],
     #[#o063, 500.0, "three"],
     #[#o064, 500.0, "four"],
     #[#o065, 500.0, "five"],
     #[#o066, 500.0, "six"],
     #[#o067, 500.0, "seven"],
     #[#o070, 500.0, "eight"],
     #[#o071, 500.0, "nine"],
     #[#o072, 278.0, "colon"],
     #[#o073, 278.0, "semicolon"],
     #[#o074, 564.0, "less"],
     #[#o075, 564.0, "equal"],
     #[#o076, 564.0, "greater"],
     #[#o077, 444.0, "question"],
     #[#o100, 921.0, "at"],
     #[#o101, 722.0, "A"],
     #[#o102, 667.0, "B"],
     #[#o103, 667.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 611.0, "E"],
     #[#o106, 556.0, "F"],
     #[#o107, 722.0, "G"],
     #[#o110, 722.0, "H"],
     #[#o111, 333.0, "I"],
     #[#o112, 389.0, "J"],
     #[#o113, 722.0, "K"],
     #[#o114, 611.0, "L"],
     #[#o115, 889.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 722.0, "O"],
     #[#o120, 556.0, "P"],
     #[#o121, 722.0, "Q"],
     #[#o122, 667.0, "R"],
     #[#o123, 556.0, "S"],
     #[#o124, 611.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 722.0, "V"],
     #[#o127, 944.0, "W"],
     #[#o130, 722.0, "X"],
     #[#o131, 722.0, "Y"],
     #[#o132, 611.0, "Z"],
     #[#o133, 333.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 333.0, "bracketright"],
     #[#o136, 469.0, "asciicircum"],
     #[#o137, 500.0, "underscore"],
     #[#o140, 333.0, "quoteleft"],
     #[#o141, 444.0, "a"],
     #[#o142, 500.0, "b"],
     #[#o143, 444.0, "c"],
     #[#o144, 500.0, "d"],
     #[#o145, 444.0, "e"],
     #[#o146, 333.0, "f"],
     #[#o147, 500.0, "g"],
     #[#o150, 500.0, "h"],
     #[#o151, 278.0, "i"],
     #[#o152, 278.0, "j"],
     #[#o153, 500.0, "k"],
     #[#o154, 278.0, "l"],
     #[#o155, 778.0, "m"],
     #[#o156, 500.0, "n"],
     #[#o157, 500.0, "o"],
     #[#o160, 500.0, "p"],
     #[#o161, 500.0, "q"],
     #[#o162, 333.0, "r"],
     #[#o163, 389.0, "s"],
     #[#o164, 278.0, "t"],
     #[#o165, 500.0, "u"],
     #[#o166, 500.0, "v"],
     #[#o167, 722.0, "w"],
     #[#o170, 500.0, "x"],
     #[#o171, 500.0, "y"],
     #[#o172, 444.0, "z"],
     #[#o173, 480.0, "braceleft"],
     #[#o174, 200.0, "bar"],
     #[#o175, 480.0, "braceright"],
     #[#o176, 541.0, "asciitilde"],
     #[#o241, 333.0, "exclamdown"],
     #[#o242, 500.0, "cent"],
     #[#o243, 500.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 500.0, "yen"],
     #[#o246, 500.0, "florin"],
     #[#o247, 500.0, "section"],
     #[#o250, 500.0, "currency"],
     #[#o251, 180.0, "quotesingle"],
     #[#o252, 444.0, "quotedblleft"],
     #[#o253, 500.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 556.0, "fi"],
     #[#o257, 556.0, "fl"],
     #[#o261, 500.0, "endash"],
     #[#o262, 500.0, "dagger"],
     #[#o263, 500.0, "daggerdbl"],
     #[#o264, 250.0, "periodcentered"],
     #[#o266, 453.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 333.0, "quotesinglbase"],
     #[#o271, 444.0, "quotedblbase"],
     #[#o272, 444.0, "quotedblright"],
     #[#o273, 500.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 444.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 889.0, "AE"],
     #[#o343, 276.0, "ordfeminine"],
     #[#o350, 611.0, "Lslash"],
     #[#o351, 722.0, "Oslash"],
     #[#o352, 889.0, "OE"],
     #[#o353, 310.0, "ordmasculine"],
     #[#o361, 667.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 278.0, "lslash"],
     #[#o371, 500.0, "oslash"],
     #[#o372, 722.0, "oe"],
     #[#o373, 500.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Times-Italic", #"serif", #"normal", #"italic",
   1000, #f, #[-176, -252, 998, 930],
   #[#[#o040, 250.0, "space"],
     #[#o041, 333.0, "exclam"],
     #[#o042, 420.0, "quotedbl"],
     #[#o043, 500.0, "numbersign"],
     #[#o044, 500.0, "dollar"],
     #[#o045, 833.0, "percent"],
     #[#o046, 778.0, "ampersand"],
     #[#o047, 333.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 500.0, "asterisk"],
     #[#o053, 675.0, "plus"],
     #[#o054, 250.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 250.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 500.0, "zero"],
     #[#o061, 500.0, "one"],
     #[#o062, 500.0, "two"],
     #[#o063, 500.0, "three"],
     #[#o064, 500.0, "four"],
     #[#o065, 500.0, "five"],
     #[#o066, 500.0, "six"],
     #[#o067, 500.0, "seven"],
     #[#o070, 500.0, "eight"],
     #[#o071, 500.0, "nine"],
     #[#o072, 333.0, "colon"],
     #[#o073, 333.0, "semicolon"],
     #[#o074, 675.0, "less"],
     #[#o075, 675.0, "equal"],
     #[#o076, 675.0, "greater"],
     #[#o077, 500.0, "question"],
     #[#o100, 920.0, "at"],
     #[#o101, 611.0, "A"],
     #[#o102, 611.0, "B"],
     #[#o103, 667.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 611.0, "E"],
     #[#o106, 611.0, "F"],
     #[#o107, 722.0, "G"],
     #[#o110, 722.0, "H"],
     #[#o111, 333.0, "I"],
     #[#o112, 444.0, "J"],
     #[#o113, 667.0, "K"],
     #[#o114, 556.0, "L"],
     #[#o115, 833.0, "M"],
     #[#o116, 667.0, "N"],
     #[#o117, 722.0, "O"],
     #[#o120, 611.0, "P"],
     #[#o121, 722.0, "Q"],
     #[#o122, 611.0, "R"],
     #[#o123, 500.0, "S"],
     #[#o124, 556.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 611.0, "V"],
     #[#o127, 833.0, "W"],
     #[#o130, 611.0, "X"],
     #[#o131, 556.0, "Y"],
     #[#o132, 556.0, "Z"],
     #[#o133, 389.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 389.0, "bracketright"],
     #[#o136, 422.0, "asciicircum"],
     #[#o137, 500.0, "underscore"],
     #[#o140, 333.0, "quoteleft"],
     #[#o141, 500.0, "a"],
     #[#o142, 500.0, "b"],
     #[#o143, 444.0, "c"],
     #[#o144, 500.0, "d"],
     #[#o145, 444.0, "e"],
     #[#o146, 278.0, "f"],
     #[#o147, 500.0, "g"],
     #[#o150, 500.0, "h"],
     #[#o151, 278.0, "i"],
     #[#o152, 278.0, "j"],
     #[#o153, 444.0, "k"],
     #[#o154, 278.0, "l"],
     #[#o155, 722.0, "m"],
     #[#o156, 500.0, "n"],
     #[#o157, 500.0, "o"],
     #[#o160, 500.0, "p"],
     #[#o161, 500.0, "q"],
     #[#o162, 389.0, "r"],
     #[#o163, 389.0, "s"],
     #[#o164, 278.0, "t"],
     #[#o165, 500.0, "u"],
     #[#o166, 444.0, "v"],
     #[#o167, 667.0, "w"],
     #[#o170, 444.0, "x"],
     #[#o171, 444.0, "y"],
     #[#o172, 389.0, "z"],
     #[#o173, 400.0, "braceleft"],
     #[#o174, 275.0, "bar"],
     #[#o175, 400.0, "braceright"],
     #[#o176, 541.0, "asciitilde"],
     #[#o241, 389.0, "exclamdown"],
     #[#o242, 500.0, "cent"],
     #[#o243, 500.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 500.0, "yen"],
     #[#o246, 500.0, "florin"],
     #[#o247, 500.0, "section"],
     #[#o250, 500.0, "currency"],
     #[#o251, 214.0, "quotesingle"],
     #[#o252, 556.0, "quotedblleft"],
     #[#o253, 500.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 500.0, "fi"],
     #[#o257, 500.0, "fl"],
     #[#o261, 500.0, "endash"],
     #[#o262, 500.0, "dagger"],
     #[#o263, 500.0, "daggerdbl"],
     #[#o264, 250.0, "periodcentered"],
     #[#o266, 523.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 333.0, "quotesinglbase"],
     #[#o271, 556.0, "quotedblbase"],
     #[#o272, 556.0, "quotedblright"],
     #[#o273, 500.0, "guillemotright"],
     #[#o274, 889.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 500.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 889.0, "emdash"],
     #[#o341, 889.0, "AE"],
     #[#o343, 276.0, "ordfeminine"],
     #[#o350, 556.0, "Lslash"],
     #[#o351, 722.0, "Oslash"],
     #[#o352, 944.0, "OE"],
     #[#o353, 310.0, "ordmasculine"],
     #[#o361, 667.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 278.0, "lslash"],
     #[#o371, 500.0, "oslash"],
     #[#o372, 667.0, "oe"],
     #[#o373, 500.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Times-Bold", #"serif", #"bold", #"roman",
   1000, #f, #[-172, -257, 1008, 965],
   #[#[#o040, 250.0, "space"],
     #[#o041, 333.0, "exclam"],
     #[#o042, 555.0, "quotedbl"],
     #[#o043, 500.0, "numbersign"],
     #[#o044, 500.0, "dollar"],
     #[#o045, 1000.0, "percent"],
     #[#o046, 833.0, "ampersand"],
     #[#o047, 333.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 500.0, "asterisk"],
     #[#o053, 570.0, "plus"],
     #[#o054, 250.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 250.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 500.0, "zero"],
     #[#o061, 500.0, "one"],
     #[#o062, 500.0, "two"],
     #[#o063, 500.0, "three"],
     #[#o064, 500.0, "four"],
     #[#o065, 500.0, "five"],
     #[#o066, 500.0, "six"],
     #[#o067, 500.0, "seven"],
     #[#o070, 500.0, "eight"],
     #[#o071, 500.0, "nine"],
     #[#o072, 333.0, "colon"],
     #[#o073, 333.0, "semicolon"],
     #[#o074, 570.0, "less"],
     #[#o075, 570.0, "equal"],
     #[#o076, 570.0, "greater"],
     #[#o077, 500.0, "question"],
     #[#o100, 930.0, "at"],
     #[#o101, 722.0, "A"],
     #[#o102, 667.0, "B"],
     #[#o103, 722.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 667.0, "E"],
     #[#o106, 611.0, "F"],
     #[#o107, 778.0, "G"],
     #[#o110, 778.0, "H"],
     #[#o111, 389.0, "I"],
     #[#o112, 500.0, "J"],
     #[#o113, 778.0, "K"],
     #[#o114, 667.0, "L"],
     #[#o115, 944.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 778.0, "O"],
     #[#o120, 611.0, "P"],
     #[#o121, 778.0, "Q"],
     #[#o122, 722.0, "R"],
     #[#o123, 556.0, "S"],
     #[#o124, 667.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 722.0, "V"],
     #[#o127, 1000.0, "W"],
     #[#o130, 722.0, "X"],
     #[#o131, 722.0, "Y"],
     #[#o132, 667.0, "Z"],
     #[#o133, 333.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 333.0, "bracketright"],
     #[#o136, 581.0, "asciicircum"],
     #[#o137, 500.0, "underscore"],
     #[#o140, 333.0, "quoteleft"],
     #[#o141, 500.0, "a"],
     #[#o142, 556.0, "b"],
     #[#o143, 444.0, "c"],
     #[#o144, 556.0, "d"],
     #[#o145, 444.0, "e"],
     #[#o146, 333.0, "f"],
     #[#o147, 500.0, "g"],
     #[#o150, 556.0, "h"],
     #[#o151, 278.0, "i"],
     #[#o152, 333.0, "j"],
     #[#o153, 556.0, "k"],
     #[#o154, 278.0, "l"],
     #[#o155, 833.0, "m"],
     #[#o156, 556.0, "n"],
     #[#o157, 500.0, "o"],
     #[#o160, 556.0, "p"],
     #[#o161, 556.0, "q"],
     #[#o162, 444.0, "r"],
     #[#o163, 389.0, "s"],
     #[#o164, 333.0, "t"],
     #[#o165, 556.0, "u"],
     #[#o166, 500.0, "v"],
     #[#o167, 722.0, "w"],
     #[#o170, 500.0, "x"],
     #[#o171, 500.0, "y"],
     #[#o172, 444.0, "z"],
     #[#o173, 394.0, "braceleft"],
     #[#o174, 220.0, "bar"],
     #[#o175, 394.0, "braceright"],
     #[#o176, 520.0, "asciitilde"],
     #[#o241, 333.0, "exclamdown"],
     #[#o242, 500.0, "cent"],
     #[#o243, 500.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 500.0, "yen"],
     #[#o246, 500.0, "florin"],
     #[#o247, 500.0, "section"],
     #[#o250, 500.0, "currency"],
     #[#o251, 278.0, "quotesingle"],
     #[#o252, 500.0, "quotedblleft"],
     #[#o253, 500.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 556.0, "fi"],
     #[#o257, 556.0, "fl"],
     #[#o261, 500.0, "endash"],
     #[#o262, 500.0, "dagger"],
     #[#o263, 500.0, "daggerdbl"],
     #[#o264, 250.0, "periodcentered"],
     #[#o266, 540.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 333.0, "quotesinglbase"],
     #[#o271, 500.0, "quotedblbase"],
     #[#o272, 500.0, "quotedblright"],
     #[#o273, 500.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 500.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 1000.0, "AE"],
     #[#o343, 300.0, "ordfeminine"],
     #[#o350, 667.0, "Lslash"],
     #[#o351, 778.0, "Oslash"],
     #[#o352, 1000.0, "OE"],
     #[#o353, 330.0, "ordmasculine"],
     #[#o361, 722.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 278.0, "lslash"],
     #[#o371, 500.0, "oslash"],
     #[#o372, 722.0, "oe"],
     #[#o373, 556.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Times-BoldItalic", #"serif", #"bold", #"italic",
   1000, #f, #[-183, -250, 1004, 973],
   #[#[#o040, 250.0, "space"],
     #[#o041, 389.0, "exclam"],
     #[#o042, 555.0, "quotedbl"],
     #[#o043, 500.0, "numbersign"],
     #[#o044, 500.0, "dollar"],
     #[#o045, 833.0, "percent"],
     #[#o046, 778.0, "ampersand"],
     #[#o047, 333.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 500.0, "asterisk"],
     #[#o053, 570.0, "plus"],
     #[#o054, 250.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 250.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 500.0, "zero"],
     #[#o061, 500.0, "one"],
     #[#o062, 500.0, "two"],
     #[#o063, 500.0, "three"],
     #[#o064, 500.0, "four"],
     #[#o065, 500.0, "five"],
     #[#o066, 500.0, "six"],
     #[#o067, 500.0, "seven"],
     #[#o070, 500.0, "eight"],
     #[#o071, 500.0, "nine"],
     #[#o072, 333.0, "colon"],
     #[#o073, 333.0, "semicolon"],
     #[#o074, 570.0, "less"],
     #[#o075, 570.0, "equal"],
     #[#o076, 570.0, "greater"],
     #[#o077, 500.0, "question"],
     #[#o100, 832.0, "at"],
     #[#o101, 667.0, "A"],
     #[#o102, 667.0, "B"],
     #[#o103, 667.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 667.0, "E"],
     #[#o106, 667.0, "F"],
     #[#o107, 722.0, "G"],
     #[#o110, 778.0, "H"],
     #[#o111, 389.0, "I"],
     #[#o112, 500.0, "J"],
     #[#o113, 667.0, "K"],
     #[#o114, 611.0, "L"],
     #[#o115, 889.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 722.0, "O"],
     #[#o120, 611.0, "P"],
     #[#o121, 722.0, "Q"],
     #[#o122, 667.0, "R"],
     #[#o123, 556.0, "S"],
     #[#o124, 611.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 667.0, "V"],
     #[#o127, 889.0, "W"],
     #[#o130, 667.0, "X"],
     #[#o131, 611.0, "Y"],
     #[#o132, 611.0, "Z"],
     #[#o133, 333.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 333.0, "bracketright"],
     #[#o136, 570.0, "asciicircum"],
     #[#o137, 500.0, "underscore"],
     #[#o140, 333.0, "quoteleft"],
     #[#o141, 500.0, "a"],
     #[#o142, 500.0, "b"],
     #[#o143, 444.0, "c"],
     #[#o144, 500.0, "d"],
     #[#o145, 444.0, "e"],
     #[#o146, 333.0, "f"],
     #[#o147, 500.0, "g"],
     #[#o150, 556.0, "h"],
     #[#o151, 278.0, "i"],
     #[#o152, 278.0, "j"],
     #[#o153, 500.0, "k"],
     #[#o154, 278.0, "l"],
     #[#o155, 778.0, "m"],
     #[#o156, 556.0, "n"],
     #[#o157, 500.0, "o"],
     #[#o160, 500.0, "p"],
     #[#o161, 500.0, "q"],
     #[#o162, 389.0, "r"],
     #[#o163, 389.0, "s"],
     #[#o164, 278.0, "t"],
     #[#o165, 556.0, "u"],
     #[#o166, 444.0, "v"],
     #[#o167, 667.0, "w"],
     #[#o170, 500.0, "x"],
     #[#o171, 444.0, "y"],
     #[#o172, 389.0, "z"],
     #[#o173, 348.0, "braceleft"],
     #[#o174, 220.0, "bar"],
     #[#o175, 348.0, "braceright"],
     #[#o176, 570.0, "asciitilde"],
     #[#o241, 389.0, "exclamdown"],
     #[#o242, 500.0, "cent"],
     #[#o243, 500.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 500.0, "yen"],
     #[#o246, 500.0, "florin"],
     #[#o247, 500.0, "section"],
     #[#o250, 500.0, "currency"],
     #[#o251, 278.0, "quotesingle"],
     #[#o252, 500.0, "quotedblleft"],
     #[#o253, 500.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 556.0, "fi"],
     #[#o257, 556.0, "fl"],
     #[#o261, 500.0, "endash"],
     #[#o262, 500.0, "dagger"],
     #[#o263, 500.0, "daggerdbl"],
     #[#o264, 250.0, "periodcentered"],
     #[#o266, 500.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 333.0, "quotesinglbase"],
     #[#o271, 500.0, "quotedblbase"],
     #[#o272, 500.0, "quotedblright"],
     #[#o273, 500.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 500.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 944.0, "AE"],
     #[#o343, 266.0, "ordfeminine"],
     #[#o350, 611.0, "Lslash"],
     #[#o351, 722.0, "Oslash"],
     #[#o352, 944.0, "OE"],
     #[#o353, 300.0, "ordmasculine"],
     #[#o361, 722.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 278.0, "lslash"],
     #[#o371, 500.0, "oslash"],
     #[#o372, 722.0, "oe"],
     #[#o373, 500.0, "germandbls"]]);


setup-laserwriter-metrics
  ("Helvetica", #"sans-serif", #"normal", #"roman",
   1000, #f, #[-174, -234, 1001, 941],
   #[#[#o040, 278.0, "space"],
     #[#o041, 278.0, "exclam"],
     #[#o042, 355.0, "quotedbl"],
     #[#o043, 556.0, "numbersign"],
     #[#o044, 556.0, "dollar"],
     #[#o045, 889.0, "percent"],
     #[#o046, 667.0, "ampersand"],
     #[#o047, 222.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 389.0, "asterisk"],
     #[#o053, 584.0, "plus"],
     #[#o054, 278.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 278.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 556.0, "zero"],
     #[#o061, 556.0, "one"],
     #[#o062, 556.0, "two"],
     #[#o063, 556.0, "three"],
     #[#o064, 556.0, "four"],
     #[#o065, 556.0, "five"],
     #[#o066, 556.0, "six"],
     #[#o067, 556.0, "seven"],
     #[#o070, 556.0, "eight"],
     #[#o071, 556.0, "nine"],
     #[#o072, 278.0, "colon"],
     #[#o073, 278.0, "semicolon"],
     #[#o074, 584.0, "less"],
     #[#o075, 584.0, "equal"],
     #[#o076, 584.0, "greater"],
     #[#o077, 556.0, "question"],
     #[#o100, 1015.0, "at"],
     #[#o101, 667.0, "A"],
     #[#o102, 667.0, "B"],
     #[#o103, 722.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 667.0, "E"],
     #[#o106, 611.0, "F"],
     #[#o107, 778.0, "G"],
     #[#o110, 722.0, "H"],
     #[#o111, 278.0, "I"],
     #[#o112, 500.0, "J"],
     #[#o113, 667.0, "K"],
     #[#o114, 556.0, "L"],
     #[#o115, 833.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 778.0, "O"],
     #[#o120, 667.0, "P"],
     #[#o121, 778.0, "Q"],
     #[#o122, 722.0, "R"],
     #[#o123, 667.0, "S"],
     #[#o124, 611.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 667.0, "V"],
     #[#o127, 944.0, "W"],
     #[#o130, 667.0, "X"],
     #[#o131, 667.0, "Y"],
     #[#o132, 611.0, "Z"],
     #[#o133, 278.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 278.0, "bracketright"],
     #[#o136, 469.0, "asciicircum"],
     #[#o137, 556.0, "underscore"],
     #[#o140, 222.0, "quoteleft"],
     #[#o141, 556.0, "a"],
     #[#o142, 556.0, "b"],
     #[#o143, 500.0, "c"],
     #[#o144, 556.0, "d"],
     #[#o145, 556.0, "e"],
     #[#o146, 278.0, "f"],
     #[#o147, 556.0, "g"],
     #[#o150, 556.0, "h"],
     #[#o151, 222.0, "i"],
     #[#o152, 222.0, "j"],
     #[#o153, 500.0, "k"],
     #[#o154, 222.0, "l"],
     #[#o155, 833.0, "m"],
     #[#o156, 556.0, "n"],
     #[#o157, 556.0, "o"],
     #[#o160, 556.0, "p"],
     #[#o161, 556.0, "q"],
     #[#o162, 333.0, "r"],
     #[#o163, 500.0, "s"],
     #[#o164, 278.0, "t"],
     #[#o165, 556.0, "u"],
     #[#o166, 500.0, "v"],
     #[#o167, 722.0, "w"],
     #[#o170, 500.0, "x"],
     #[#o171, 500.0, "y"],
     #[#o172, 500.0, "z"],
     #[#o173, 334.0, "braceleft"],
     #[#o174, 260.0, "bar"],
     #[#o175, 334.0, "braceright"],
     #[#o176, 584.0, "asciitilde"],
     #[#o241, 333.0, "exclamdown"],
     #[#o242, 556.0, "cent"],
     #[#o243, 556.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 556.0, "yen"],
     #[#o246, 556.0, "florin"],
     #[#o247, 556.0, "section"],
     #[#o250, 556.0, "currency"],
     #[#o251, 191.0, "quotesingle"],
     #[#o252, 333.0, "quotedblleft"],
     #[#o253, 556.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 500.0, "fi"],
     #[#o257, 500.0, "fl"],
     #[#o261, 556.0, "endash"],
     #[#o262, 556.0, "dagger"],
     #[#o263, 556.0, "daggerdbl"],
     #[#o264, 278.0, "periodcentered"],
     #[#o266, 537.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 222.0, "quotesinglbase"],
     #[#o271, 333.0, "quotedblbase"],
     #[#o272, 333.0, "quotedblright"],
     #[#o273, 556.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 611.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 1000.0, "AE"],
     #[#o343, 370.0, "ordfeminine"],
     #[#o350, 556.0, "Lslash"],
     #[#o351, 778.0, "Oslash"],
     #[#o352, 1000.0, "OE"],
     #[#o353, 365.0, "ordmasculine"],
     #[#o361, 889.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 222.0, "lslash"],
     #[#o371, 611.0, "oslash"],
     #[#o372, 944.0, "oe"],
     #[#o373, 611.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Helvetica-Oblique", #"sans-serif", #"normal", #"italic",
   1000, #f, #[-174, -234, 1001, 941],
   #[#[#o040, 278.0, "space"],
     #[#o041, 278.0, "exclam"],
     #[#o042, 355.0, "quotedbl"],
     #[#o043, 556.0, "numbersign"],
     #[#o044, 556.0, "dollar"],
     #[#o045, 889.0, "percent"],
     #[#o046, 667.0, "ampersand"],
     #[#o047, 222.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 389.0, "asterisk"],
     #[#o053, 584.0, "plus"],
     #[#o054, 278.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 278.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 556.0, "zero"],
     #[#o061, 556.0, "one"],
     #[#o062, 556.0, "two"],
     #[#o063, 556.0, "three"],
     #[#o064, 556.0, "four"],
     #[#o065, 556.0, "five"],
     #[#o066, 556.0, "six"],
     #[#o067, 556.0, "seven"],
     #[#o070, 556.0, "eight"],
     #[#o071, 556.0, "nine"],
     #[#o072, 278.0, "colon"],
     #[#o073, 278.0, "semicolon"],
     #[#o074, 584.0, "less"],
     #[#o075, 584.0, "equal"],
     #[#o076, 584.0, "greater"],
     #[#o077, 556.0, "question"],
     #[#o100, 1015.0, "at"],
     #[#o101, 667.0, "A"],
     #[#o102, 667.0, "B"],
     #[#o103, 722.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 667.0, "E"],
     #[#o106, 611.0, "F"],
     #[#o107, 778.0, "G"],
     #[#o110, 722.0, "H"],
     #[#o111, 278.0, "I"],
     #[#o112, 500.0, "J"],
     #[#o113, 667.0, "K"],
     #[#o114, 556.0, "L"],
     #[#o115, 833.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 778.0, "O"],
     #[#o120, 667.0, "P"],
     #[#o121, 778.0, "Q"],
     #[#o122, 722.0, "R"],
     #[#o123, 667.0, "S"],
     #[#o124, 611.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 667.0, "V"],
     #[#o127, 944.0, "W"],
     #[#o130, 667.0, "X"],
     #[#o131, 667.0, "Y"],
     #[#o132, 611.0, "Z"],
     #[#o133, 278.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 278.0, "bracketright"],
     #[#o136, 469.0, "asciicircum"],
     #[#o137, 556.0, "underscore"],
     #[#o140, 222.0, "quoteleft"],
     #[#o141, 556.0, "a"],
     #[#o142, 556.0, "b"],
     #[#o143, 500.0, "c"],
     #[#o144, 556.0, "d"],
     #[#o145, 556.0, "e"],
     #[#o146, 278.0, "f"],
     #[#o147, 556.0, "g"],
     #[#o150, 556.0, "h"],
     #[#o151, 222.0, "i"],
     #[#o152, 222.0, "j"],
     #[#o153, 500.0, "k"],
     #[#o154, 222.0, "l"],
     #[#o155, 833.0, "m"],
     #[#o156, 556.0, "n"],
     #[#o157, 556.0, "o"],
     #[#o160, 556.0, "p"],
     #[#o161, 556.0, "q"],
     #[#o162, 333.0, "r"],
     #[#o163, 500.0, "s"],
     #[#o164, 278.0, "t"],
     #[#o165, 556.0, "u"],
     #[#o166, 500.0, "v"],
     #[#o167, 722.0, "w"],
     #[#o170, 500.0, "x"],
     #[#o171, 500.0, "y"],
     #[#o172, 500.0, "z"],
     #[#o173, 334.0, "braceleft"],
     #[#o174, 260.0, "bar"],
     #[#o175, 334.0, "braceright"],
     #[#o176, 584.0, "asciitilde"],
     #[#o241, 333.0, "exclamdown"],
     #[#o242, 556.0, "cent"],
     #[#o243, 556.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 556.0, "yen"],
     #[#o246, 556.0, "florin"],
     #[#o247, 556.0, "section"],
     #[#o250, 556.0, "currency"],
     #[#o251, 191.0, "quotesingle"],
     #[#o252, 333.0, "quotedblleft"],
     #[#o253, 556.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 500.0, "fi"],
     #[#o257, 500.0, "fl"],
     #[#o261, 556.0, "endash"],
     #[#o262, 556.0, "dagger"],
     #[#o263, 556.0, "daggerdbl"],
     #[#o264, 278.0, "periodcentered"],
     #[#o266, 537.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 222.0, "quotesinglbase"],
     #[#o271, 333.0, "quotedblbase"],
     #[#o272, 333.0, "quotedblright"],
     #[#o273, 556.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 611.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 1000.0, "AE"],
     #[#o343, 370.0, "ordfeminine"],
     #[#o350, 556.0, "Lslash"],
     #[#o351, 778.0, "Oslash"],
     #[#o352, 1000.0, "OE"],
     #[#o353, 365.0, "ordmasculine"],
     #[#o361, 889.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 222.0, "lslash"],
     #[#o371, 611.0, "oslash"],
     #[#o372, 944.0, "oe"],
     #[#o373, 611.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Helvetica-Bold", #"sans-serif", #"bold", #"roman",
   1000, #f, #[-173, -241, 1003, 936],
   #[#[#o040, 278.0, "space"],
     #[#o041, 333.0, "exclam"],
     #[#o042, 474.0, "quotedbl"],
     #[#o043, 556.0, "numbersign"],
     #[#o044, 556.0, "dollar"],
     #[#o045, 889.0, "percent"],
     #[#o046, 722.0, "ampersand"],
     #[#o047, 278.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 389.0, "asterisk"],
     #[#o053, 584.0, "plus"],
     #[#o054, 278.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 278.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 556.0, "zero"],
     #[#o061, 556.0, "one"],
     #[#o062, 556.0, "two"],
     #[#o063, 556.0, "three"],
     #[#o064, 556.0, "four"],
     #[#o065, 556.0, "five"],
     #[#o066, 556.0, "six"],
     #[#o067, 556.0, "seven"],
     #[#o070, 556.0, "eight"],
     #[#o071, 556.0, "nine"],
     #[#o072, 333.0, "colon"],
     #[#o073, 333.0, "semicolon"],
     #[#o074, 584.0, "less"],
     #[#o075, 584.0, "equal"],
     #[#o076, 584.0, "greater"],
     #[#o077, 611.0, "question"],
     #[#o100, 975.0, "at"],
     #[#o101, 722.0, "A"],
     #[#o102, 722.0, "B"],
     #[#o103, 722.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 667.0, "E"],
     #[#o106, 611.0, "F"],
     #[#o107, 778.0, "G"],
     #[#o110, 722.0, "H"],
     #[#o111, 278.0, "I"],
     #[#o112, 556.0, "J"],
     #[#o113, 722.0, "K"],
     #[#o114, 611.0, "L"],
     #[#o115, 833.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 778.0, "O"],
     #[#o120, 667.0, "P"],
     #[#o121, 778.0, "Q"],
     #[#o122, 722.0, "R"],
     #[#o123, 667.0, "S"],
     #[#o124, 611.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 667.0, "V"],
     #[#o127, 944.0, "W"],
     #[#o130, 667.0, "X"],
     #[#o131, 667.0, "Y"],
     #[#o132, 611.0, "Z"],
     #[#o133, 333.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 333.0, "bracketright"],
     #[#o136, 584.0, "asciicircum"],
     #[#o137, 556.0, "underscore"],
     #[#o140, 278.0, "quoteleft"],
     #[#o141, 556.0, "a"],
     #[#o142, 611.0, "b"],
     #[#o143, 556.0, "c"],
     #[#o144, 611.0, "d"],
     #[#o145, 556.0, "e"],
     #[#o146, 333.0, "f"],
     #[#o147, 611.0, "g"],
     #[#o150, 611.0, "h"],
     #[#o151, 278.0, "i"],
     #[#o152, 278.0, "j"],
     #[#o153, 556.0, "k"],
     #[#o154, 278.0, "l"],
     #[#o155, 889.0, "m"],
     #[#o156, 611.0, "n"],
     #[#o157, 611.0, "o"],
     #[#o160, 611.0, "p"],
     #[#o161, 611.0, "q"],
     #[#o162, 389.0, "r"],
     #[#o163, 556.0, "s"],
     #[#o164, 333.0, "t"],
     #[#o165, 611.0, "u"],
     #[#o166, 556.0, "v"],
     #[#o167, 778.0, "w"],
     #[#o170, 556.0, "x"],
     #[#o171, 556.0, "y"],
     #[#o172, 500.0, "z"],
     #[#o173, 389.0, "braceleft"],
     #[#o174, 280.0, "bar"],
     #[#o175, 389.0, "braceright"],
     #[#o176, 584.0, "asciitilde"],
     #[#o241, 333.0, "exclamdown"],
     #[#o242, 556.0, "cent"],
     #[#o243, 556.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 556.0, "yen"],
     #[#o246, 556.0, "florin"],
     #[#o247, 556.0, "section"],
     #[#o250, 556.0, "currency"],
     #[#o251, 238.0, "quotesingle"],
     #[#o252, 500.0, "quotedblleft"],
     #[#o253, 556.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 611.0, "fi"],
     #[#o257, 611.0, "fl"],
     #[#o261, 556.0, "endash"],
     #[#o262, 556.0, "dagger"],
     #[#o263, 556.0, "daggerdbl"],
     #[#o264, 278.0, "periodcentered"],
     #[#o266, 556.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 278.0, "quotesinglbase"],
     #[#o271, 500.0, "quotedblbase"],
     #[#o272, 500.0, "quotedblright"],
     #[#o273, 556.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 611.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 1000.0, "AE"],
     #[#o343, 370.0, "ordfeminine"],
     #[#o350, 611.0, "Lslash"],
     #[#o351, 778.0, "Oslash"],
     #[#o352, 1000.0, "OE"],
     #[#o353, 365.0, "ordmasculine"],
     #[#o361, 889.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 278.0, "lslash"],
     #[#o371, 611.0, "oslash"],
     #[#o372, 944.0, "oe"],
     #[#o373, 611.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Helvetica-BoldOblique", #"sans-serif", #"bold", #"italic",
   1000, #f, #[-173, -241, 1003, 936],
   #[#[#o040, 278.0, "space"],
     #[#o041, 333.0, "exclam"],
     #[#o042, 474.0, "quotedbl"],
     #[#o043, 556.0, "numbersign"],
     #[#o044, 556.0, "dollar"],
     #[#o045, 889.0, "percent"],
     #[#o046, 722.0, "ampersand"],
     #[#o047, 278.0, "quoteright"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 389.0, "asterisk"],
     #[#o053, 584.0, "plus"],
     #[#o054, 278.0, "comma"],
     #[#o055, 333.0, "hyphen"],
     #[#o056, 278.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 556.0, "zero"],
     #[#o061, 556.0, "one"],
     #[#o062, 556.0, "two"],
     #[#o063, 556.0, "three"],
     #[#o064, 556.0, "four"],
     #[#o065, 556.0, "five"],
     #[#o066, 556.0, "six"],
     #[#o067, 556.0, "seven"],
     #[#o070, 556.0, "eight"],
     #[#o071, 556.0, "nine"],
     #[#o072, 333.0, "colon"],
     #[#o073, 333.0, "semicolon"],
     #[#o074, 584.0, "less"],
     #[#o075, 584.0, "equal"],
     #[#o076, 584.0, "greater"],
     #[#o077, 611.0, "question"],
     #[#o100, 975.0, "at"],
     #[#o101, 722.0, "A"],
     #[#o102, 722.0, "B"],
     #[#o103, 722.0, "C"],
     #[#o104, 722.0, "D"],
     #[#o105, 667.0, "E"],
     #[#o106, 611.0, "F"],
     #[#o107, 778.0, "G"],
     #[#o110, 722.0, "H"],
     #[#o111, 278.0, "I"],
     #[#o112, 556.0, "J"],
     #[#o113, 722.0, "K"],
     #[#o114, 611.0, "L"],
     #[#o115, 833.0, "M"],
     #[#o116, 722.0, "N"],
     #[#o117, 778.0, "O"],
     #[#o120, 667.0, "P"],
     #[#o121, 778.0, "Q"],
     #[#o122, 722.0, "R"],
     #[#o123, 667.0, "S"],
     #[#o124, 611.0, "T"],
     #[#o125, 722.0, "U"],
     #[#o126, 667.0, "V"],
     #[#o127, 944.0, "W"],
     #[#o130, 667.0, "X"],
     #[#o131, 667.0, "Y"],
     #[#o132, 611.0, "Z"],
     #[#o133, 333.0, "bracketleft"],
     #[#o134, 278.0, "backslash"],
     #[#o135, 333.0, "bracketright"],
     #[#o136, 584.0, "asciicircum"],
     #[#o137, 556.0, "underscore"],
     #[#o140, 278.0, "quoteleft"],
     #[#o141, 556.0, "a"],
     #[#o142, 611.0, "b"],
     #[#o143, 556.0, "c"],
     #[#o144, 611.0, "d"],
     #[#o145, 556.0, "e"],
     #[#o146, 333.0, "f"],
     #[#o147, 611.0, "g"],
     #[#o150, 611.0, "h"],
     #[#o151, 278.0, "i"],
     #[#o152, 278.0, "j"],
     #[#o153, 556.0, "k"],
     #[#o154, 278.0, "l"],
     #[#o155, 889.0, "m"],
     #[#o156, 611.0, "n"],
     #[#o157, 611.0, "o"],
     #[#o160, 611.0, "p"],
     #[#o161, 611.0, "q"],
     #[#o162, 389.0, "r"],
     #[#o163, 556.0, "s"],
     #[#o164, 333.0, "t"],
     #[#o165, 611.0, "u"],
     #[#o166, 556.0, "v"],
     #[#o167, 778.0, "w"],
     #[#o170, 556.0, "x"],
     #[#o171, 556.0, "y"],
     #[#o172, 500.0, "z"],
     #[#o173, 389.0, "braceleft"],
     #[#o174, 280.0, "bar"],
     #[#o175, 389.0, "braceright"],
     #[#o176, 584.0, "asciitilde"],
     #[#o241, 333.0, "exclamdown"],
     #[#o242, 556.0, "cent"],
     #[#o243, 556.0, "sterling"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 556.0, "yen"],
     #[#o246, 556.0, "florin"],
     #[#o247, 556.0, "section"],
     #[#o250, 556.0, "currency"],
     #[#o251, 238.0, "quotesingle"],
     #[#o252, 500.0, "quotedblleft"],
     #[#o253, 556.0, "guillemotleft"],
     #[#o254, 333.0, "guilsinglleft"],
     #[#o255, 333.0, "guilsinglright"],
     #[#o256, 611.0, "fi"],
     #[#o257, 611.0, "fl"],
     #[#o261, 556.0, "endash"],
     #[#o262, 556.0, "dagger"],
     #[#o263, 556.0, "daggerdbl"],
     #[#o264, 278.0, "periodcentered"],
     #[#o266, 556.0, "paragraph"],
     #[#o267, 350.0, "bullet"],
     #[#o270, 278.0, "quotesinglbase"],
     #[#o271, 500.0, "quotedblbase"],
     #[#o272, 500.0, "quotedblright"],
     #[#o273, 556.0, "guillemotright"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 1000.0, "perthousand"],
     #[#o277, 611.0, "questiondown"],
     #[#o301, 333.0, "grave"],
     #[#o302, 333.0, "acute"],
     #[#o303, 333.0, "circumflex"],
     #[#o304, 333.0, "tilde"],
     #[#o305, 333.0, "macron"],
     #[#o306, 333.0, "breve"],
     #[#o307, 333.0, "dotaccent"],
     #[#o310, 333.0, "dieresis"],
     #[#o312, 333.0, "ring"],
     #[#o313, 333.0, "cedilla"],
     #[#o315, 333.0, "hungarumlaut"],
     #[#o316, 333.0, "ogonek"],
     #[#o317, 333.0, "caron"],
     #[#o320, 1000.0, "emdash"],
     #[#o341, 1000.0, "AE"],
     #[#o343, 370.0, "ordfeminine"],
     #[#o350, 611.0, "Lslash"],
     #[#o351, 778.0, "Oslash"],
     #[#o352, 1000.0, "OE"],
     #[#o353, 365.0, "ordmasculine"],
     #[#o361, 889.0, "ae"],
     #[#o365, 278.0, "dotlessi"],
     #[#o370, 278.0, "lslash"],
     #[#o371, 611.0, "oslash"],
     #[#o372, 944.0, "oe"],
     #[#o373, 611.0, "germandbls"]]);


// As it happens, the Courier metrics never get used because they are fixed-width
// fonts that get handled specially by the PostScript code itself.

setup-laserwriter-metrics
  ("Courier", #"fix", #"normal", #"roman",
   1000, 0.6, #[0, -251, 600, 700],
   #[#[#o040, 600.0, "space"],
     #[#o041, 600.0, "exclam"],
     #[#o042, 600.0, "quotedbl"],
     #[#o043, 600.0, "numbersign"],
     #[#o044, 600.0, "dollar"],
     #[#o045, 600.0, "percent"],
     #[#o046, 600.0, "ampersand"],
     #[#o047, 600.0, "quoteright"],
     #[#o050, 600.0, "parenleft"],
     #[#o051, 600.0, "parenright"],
     #[#o052, 600.0, "asterisk"],
     #[#o053, 600.0, "plus"],
     #[#o054, 600.0, "comma"],
     #[#o055, 600.0, "hyphen"],
     #[#o056, 600.0, "period"],
     #[#o057, 600.0, "slash"],
     #[#o060, 600.0, "zero"],
     #[#o061, 600.0, "one"],
     #[#o062, 600.0, "two"],
     #[#o063, 600.0, "three"],
     #[#o064, 600.0, "four"],
     #[#o065, 600.0, "five"],
     #[#o066, 600.0, "six"],
     #[#o067, 600.0, "seven"],
     #[#o070, 600.0, "eight"],
     #[#o071, 600.0, "nine"],
     #[#o072, 600.0, "colon"],
     #[#o073, 600.0, "semicolon"],
     #[#o074, 600.0, "less"],
     #[#o075, 600.0, "equal"],
     #[#o076, 600.0, "greater"],
     #[#o077, 600.0, "question"],
     #[#o100, 600.0, "at"],
     #[#o101, 600.0, "A"],
     #[#o102, 600.0, "B"],
     #[#o103, 600.0, "C"],
     #[#o104, 600.0, "D"],
     #[#o105, 600.0, "E"],
     #[#o106, 600.0, "F"],
     #[#o107, 600.0, "G"],
     #[#o110, 600.0, "H"],
     #[#o111, 600.0, "I"],
     #[#o112, 600.0, "J"],
     #[#o113, 600.0, "K"],
     #[#o114, 600.0, "L"],
     #[#o115, 600.0, "M"],
     #[#o116, 600.0, "N"],
     #[#o117, 600.0, "O"],
     #[#o120, 600.0, "P"],
     #[#o121, 600.0, "Q"],
     #[#o122, 600.0, "R"],
     #[#o123, 600.0, "S"],
     #[#o124, 600.0, "T"],
     #[#o125, 600.0, "U"],
     #[#o126, 600.0, "V"],
     #[#o127, 600.0, "W"],
     #[#o130, 600.0, "X"],
     #[#o131, 600.0, "Y"],
     #[#o132, 600.0, "Z"],
     #[#o133, 600.0, "bracketleft"],
     #[#o134, 600.0, "backslash"],
     #[#o135, 600.0, "bracketright"],
     #[#o136, 600.0, "asciicircum"],
     #[#o137, 600.0, "underscore"],
     #[#o140, 600.0, "quoteleft"],
     #[#o141, 600.0, "a"],
     #[#o142, 600.0, "b"],
     #[#o143, 600.0, "c"],
     #[#o144, 600.0, "d"],
     #[#o145, 600.0, "e"],
     #[#o146, 600.0, "f"],
     #[#o147, 600.0, "g"],
     #[#o150, 600.0, "h"],
     #[#o151, 600.0, "i"],
     #[#o152, 600.0, "j"],
     #[#o153, 600.0, "k"],
     #[#o154, 600.0, "l"],
     #[#o155, 600.0, "m"],
     #[#o156, 600.0, "n"],
     #[#o157, 600.0, "o"],
     #[#o160, 600.0, "p"],
     #[#o161, 600.0, "q"],
     #[#o162, 600.0, "r"],
     #[#o163, 600.0, "s"],
     #[#o164, 600.0, "t"],
     #[#o165, 600.0, "u"],
     #[#o166, 600.0, "v"],
     #[#o167, 600.0, "w"],
     #[#o170, 600.0, "x"],
     #[#o171, 600.0, "y"],
     #[#o172, 600.0, "z"],
     #[#o173, 600.0, "braceleft"],
     #[#o174, 600.0, "bar"],
     #[#o175, 600.0, "braceright"],
     #[#o176, 600.0, "asciitilde"],
     #[#o241, 600.0, "exclamdown"],
     #[#o242, 600.0, "cent"],
     #[#o243, 600.0, "sterling"],
     #[#o244, 600.0, "fraction"],
     #[#o245, 600.0, "yen"],
     #[#o246, 600.0, "florin"],
     #[#o247, 600.0, "section"],
     #[#o250, 600.0, "currency"],
     #[#o251, 600.0, "quotesingle"],
     #[#o252, 600.0, "quotedblleft"],
     #[#o253, 600.0, "guillemotleft"],
     #[#o254, 600.0, "guilsinglleft"],
     #[#o255, 600.0, "guilsinglright"],
     #[#o256, 600.0, "fi"],
     #[#o257, 600.0, "fl"],
     #[#o261, 600.0, "endash"],
     #[#o262, 600.0, "dagger"],
     #[#o263, 600.0, "daggerdbl"],
     #[#o264, 600.0, "periodcentered"],
     #[#o266, 600.0, "paragraph"],
     #[#o267, 600.0, "bullet"],
     #[#o270, 600.0, "quotesinglbase"],
     #[#o271, 600.0, "quotedblbase"],
     #[#o272, 600.0, "quotedblright"],
     #[#o273, 600.0, "guillemotright"],
     #[#o274, 600.0, "ellipsis"],
     #[#o275, 600.0, "perthousand"],
     #[#o277, 600.0, "questiondown"],
     #[#o301, 600.0, "grave"],
     #[#o302, 600.0, "acute"],
     #[#o303, 600.0, "circumflex"],
     #[#o304, 600.0, "tilde"],
     #[#o305, 600.0, "macron"],
     #[#o306, 600.0, "breve"],
     #[#o307, 600.0, "dotaccent"],
     #[#o310, 600.0, "dieresis"],
     #[#o312, 600.0, "ring"],
     #[#o313, 600.0, "cedilla"],
     #[#o315, 600.0, "hungarumlaut"],
     #[#o316, 600.0, "ogonek"],
     #[#o317, 600.0, "caron"],
     #[#o320, 600.0, "emdash"],
     #[#o341, 600.0, "AE"],
     #[#o343, 600.0, "ordfeminine"],
     #[#o350, 600.0, "Lslash"],
     #[#o351, 600.0, "Oslash"],
     #[#o352, 600.0, "OE"],
     #[#o353, 600.0, "ordmasculine"],
     #[#o361, 600.0, "ae"],
     #[#o365, 600.0, "dotlessi"],
     #[#o370, 600.0, "lslash"],
     #[#o371, 600.0, "oslash"],
     #[#o372, 600.0, "oe"],
     #[#o373, 600.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Courier-Oblique", #"fix", #"normal", #"italic",
   1000, 0.6, #[0, -251, 600, 700],
   #[#[#o040, 600.0, "space"],
     #[#o041, 600.0, "exclam"],
     #[#o042, 600.0, "quotedbl"],
     #[#o043, 600.0, "numbersign"],
     #[#o044, 600.0, "dollar"],
     #[#o045, 600.0, "percent"],
     #[#o046, 600.0, "ampersand"],
     #[#o047, 600.0, "quoteright"],
     #[#o050, 600.0, "parenleft"],
     #[#o051, 600.0, "parenright"],
     #[#o052, 600.0, "asterisk"],
     #[#o053, 600.0, "plus"],
     #[#o054, 600.0, "comma"],
     #[#o055, 600.0, "hyphen"],
     #[#o056, 600.0, "period"],
     #[#o057, 600.0, "slash"],
     #[#o060, 600.0, "zero"],
     #[#o061, 600.0, "one"],
     #[#o062, 600.0, "two"],
     #[#o063, 600.0, "three"],
     #[#o064, 600.0, "four"],
     #[#o065, 600.0, "five"],
     #[#o066, 600.0, "six"],
     #[#o067, 600.0, "seven"],
     #[#o070, 600.0, "eight"],
     #[#o071, 600.0, "nine"],
     #[#o072, 600.0, "colon"],
     #[#o073, 600.0, "semicolon"],
     #[#o074, 600.0, "less"],
     #[#o075, 600.0, "equal"],
     #[#o076, 600.0, "greater"],
     #[#o077, 600.0, "question"],
     #[#o100, 600.0, "at"],
     #[#o101, 600.0, "A"],
     #[#o102, 600.0, "B"],
     #[#o103, 600.0, "C"],
     #[#o104, 600.0, "D"],
     #[#o105, 600.0, "E"],
     #[#o106, 600.0, "F"],
     #[#o107, 600.0, "G"],
     #[#o110, 600.0, "H"],
     #[#o111, 600.0, "I"],
     #[#o112, 600.0, "J"],
     #[#o113, 600.0, "K"],
     #[#o114, 600.0, "L"],
     #[#o115, 600.0, "M"],
     #[#o116, 600.0, "N"],
     #[#o117, 600.0, "O"],
     #[#o120, 600.0, "P"],
     #[#o121, 600.0, "Q"],
     #[#o122, 600.0, "R"],
     #[#o123, 600.0, "S"],
     #[#o124, 600.0, "T"],
     #[#o125, 600.0, "U"],
     #[#o126, 600.0, "V"],
     #[#o127, 600.0, "W"],
     #[#o130, 600.0, "X"],
     #[#o131, 600.0, "Y"],
     #[#o132, 600.0, "Z"],
     #[#o133, 600.0, "bracketleft"],
     #[#o134, 600.0, "backslash"],
     #[#o135, 600.0, "bracketright"],
     #[#o136, 600.0, "asciicircum"],
     #[#o137, 600.0, "underscore"],
     #[#o140, 600.0, "quoteleft"],
     #[#o141, 600.0, "a"],
     #[#o142, 600.0, "b"],
     #[#o143, 600.0, "c"],
     #[#o144, 600.0, "d"],
     #[#o145, 600.0, "e"],
     #[#o146, 600.0, "f"],
     #[#o147, 600.0, "g"],
     #[#o150, 600.0, "h"],
     #[#o151, 600.0, "i"],
     #[#o152, 600.0, "j"],
     #[#o153, 600.0, "k"],
     #[#o154, 600.0, "l"],
     #[#o155, 600.0, "m"],
     #[#o156, 600.0, "n"],
     #[#o157, 600.0, "o"],
     #[#o160, 600.0, "p"],
     #[#o161, 600.0, "q"],
     #[#o162, 600.0, "r"],
     #[#o163, 600.0, "s"],
     #[#o164, 600.0, "t"],
     #[#o165, 600.0, "u"],
     #[#o166, 600.0, "v"],
     #[#o167, 600.0, "w"],
     #[#o170, 600.0, "x"],
     #[#o171, 600.0, "y"],
     #[#o172, 600.0, "z"],
     #[#o173, 600.0, "braceleft"],
     #[#o174, 600.0, "bar"],
     #[#o175, 600.0, "braceright"],
     #[#o176, 600.0, "asciitilde"],
     #[#o241, 600.0, "exclamdown"],
     #[#o242, 600.0, "cent"],
     #[#o243, 600.0, "sterling"],
     #[#o244, 600.0, "fraction"],
     #[#o245, 600.0, "yen"],
     #[#o246, 600.0, "florin"],
     #[#o247, 600.0, "section"],
     #[#o250, 600.0, "currency"],
     #[#o251, 600.0, "quotesingle"],
     #[#o252, 600.0, "quotedblleft"],
     #[#o253, 600.0, "guillemotleft"],
     #[#o254, 600.0, "guilsinglleft"],
     #[#o255, 600.0, "guilsinglright"],
     #[#o256, 600.0, "fi"],
     #[#o257, 600.0, "fl"],
     #[#o261, 600.0, "endash"],
     #[#o262, 600.0, "dagger"],
     #[#o263, 600.0, "daggerdbl"],
     #[#o264, 600.0, "periodcentered"],
     #[#o266, 600.0, "paragraph"],
     #[#o267, 600.0, "bullet"],
     #[#o270, 600.0, "quotesinglbase"],
     #[#o271, 600.0, "quotedblbase"],
     #[#o272, 600.0, "quotedblright"],
     #[#o273, 600.0, "guillemotright"],
     #[#o274, 600.0, "ellipsis"],
     #[#o275, 600.0, "perthousand"],
     #[#o277, 600.0, "questiondown"],
     #[#o301, 600.0, "grave"],
     #[#o302, 600.0, "acute"],
     #[#o303, 600.0, "circumflex"],
     #[#o304, 600.0, "tilde"],
     #[#o305, 600.0, "macron"],
     #[#o306, 600.0, "breve"],
     #[#o307, 600.0, "dotaccent"],
     #[#o310, 600.0, "dieresis"],
     #[#o312, 600.0, "ring"],
     #[#o313, 600.0, "cedilla"],
     #[#o315, 600.0, "hungarumlaut"],
     #[#o316, 600.0, "ogonek"],
     #[#o317, 600.0, "caron"],
     #[#o320, 600.0, "emdash"],
     #[#o341, 600.0, "AE"],
     #[#o343, 600.0, "ordfeminine"],
     #[#o350, 600.0, "Lslash"],
     #[#o351, 600.0, "Oslash"],
     #[#o352, 600.0, "OE"],
     #[#o353, 600.0, "ordmasculine"],
     #[#o361, 600.0, "ae"],
     #[#o365, 600.0, "dotlessi"],
     #[#o370, 600.0, "lslash"],
     #[#o371, 600.0, "oslash"],
     #[#o372, 600.0, "oe"],
     #[#o373, 600.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Courier-Bold", #"fix", #"bold", #"roman",
   1000, 0.6, #[0, -251, 600, 700],
   #[#[#o040, 600.0, "space"],
     #[#o041, 600.0, "exclam"],
     #[#o042, 600.0, "quotedbl"],
     #[#o043, 600.0, "numbersign"],
     #[#o044, 600.0, "dollar"],
     #[#o045, 600.0, "percent"],
     #[#o046, 600.0, "ampersand"],
     #[#o047, 600.0, "quoteright"],
     #[#o050, 600.0, "parenleft"],
     #[#o051, 600.0, "parenright"],
     #[#o052, 600.0, "asterisk"],
     #[#o053, 600.0, "plus"],
     #[#o054, 600.0, "comma"],
     #[#o055, 600.0, "hyphen"],
     #[#o056, 600.0, "period"],
     #[#o057, 600.0, "slash"],
     #[#o060, 600.0, "zero"],
     #[#o061, 600.0, "one"],
     #[#o062, 600.0, "two"],
     #[#o063, 600.0, "three"],
     #[#o064, 600.0, "four"],
     #[#o065, 600.0, "five"],
     #[#o066, 600.0, "six"],
     #[#o067, 600.0, "seven"],
     #[#o070, 600.0, "eight"],
     #[#o071, 600.0, "nine"],
     #[#o072, 600.0, "colon"],
     #[#o073, 600.0, "semicolon"],
     #[#o074, 600.0, "less"],
     #[#o075, 600.0, "equal"],
     #[#o076, 600.0, "greater"],
     #[#o077, 600.0, "question"],
     #[#o100, 600.0, "at"],
     #[#o101, 600.0, "A"],
     #[#o102, 600.0, "B"],
     #[#o103, 600.0, "C"],
     #[#o104, 600.0, "D"],
     #[#o105, 600.0, "E"],
     #[#o106, 600.0, "F"],
     #[#o107, 600.0, "G"],
     #[#o110, 600.0, "H"],
     #[#o111, 600.0, "I"],
     #[#o112, 600.0, "J"],
     #[#o113, 600.0, "K"],
     #[#o114, 600.0, "L"],
     #[#o115, 600.0, "M"],
     #[#o116, 600.0, "N"],
     #[#o117, 600.0, "O"],
     #[#o120, 600.0, "P"],
     #[#o121, 600.0, "Q"],
     #[#o122, 600.0, "R"],
     #[#o123, 600.0, "S"],
     #[#o124, 600.0, "T"],
     #[#o125, 600.0, "U"],
     #[#o126, 600.0, "V"],
     #[#o127, 600.0, "W"],
     #[#o130, 600.0, "X"],
     #[#o131, 600.0, "Y"],
     #[#o132, 600.0, "Z"],
     #[#o133, 600.0, "bracketleft"],
     #[#o134, 600.0, "backslash"],
     #[#o135, 600.0, "bracketright"],
     #[#o136, 600.0, "asciicircum"],
     #[#o137, 600.0, "underscore"],
     #[#o140, 600.0, "quoteleft"],
     #[#o141, 600.0, "a"],
     #[#o142, 600.0, "b"],
     #[#o143, 600.0, "c"],
     #[#o144, 600.0, "d"],
     #[#o145, 600.0, "e"],
     #[#o146, 600.0, "f"],
     #[#o147, 600.0, "g"],
     #[#o150, 600.0, "h"],
     #[#o151, 600.0, "i"],
     #[#o152, 600.0, "j"],
     #[#o153, 600.0, "k"],
     #[#o154, 600.0, "l"],
     #[#o155, 600.0, "m"],
     #[#o156, 600.0, "n"],
     #[#o157, 600.0, "o"],
     #[#o160, 600.0, "p"],
     #[#o161, 600.0, "q"],
     #[#o162, 600.0, "r"],
     #[#o163, 600.0, "s"],
     #[#o164, 600.0, "t"],
     #[#o165, 600.0, "u"],
     #[#o166, 600.0, "v"],
     #[#o167, 600.0, "w"],
     #[#o170, 600.0, "x"],
     #[#o171, 600.0, "y"],
     #[#o172, 600.0, "z"],
     #[#o173, 600.0, "braceleft"],
     #[#o174, 600.0, "bar"],
     #[#o175, 600.0, "braceright"],
     #[#o176, 600.0, "asciitilde"],
     #[#o241, 600.0, "exclamdown"],
     #[#o242, 600.0, "cent"],
     #[#o243, 600.0, "sterling"],
     #[#o244, 600.0, "fraction"],
     #[#o245, 600.0, "yen"],
     #[#o246, 600.0, "florin"],
     #[#o247, 600.0, "section"],
     #[#o250, 600.0, "currency"],
     #[#o251, 600.0, "quotesingle"],
     #[#o252, 600.0, "quotedblleft"],
     #[#o253, 600.0, "guillemotleft"],
     #[#o254, 600.0, "guilsinglleft"],
     #[#o255, 600.0, "guilsinglright"],
     #[#o256, 600.0, "fi"],
     #[#o257, 600.0, "fl"],
     #[#o261, 600.0, "endash"],
     #[#o262, 600.0, "dagger"],
     #[#o263, 600.0, "daggerdbl"],
     #[#o264, 600.0, "periodcentered"],
     #[#o266, 600.0, "paragraph"],
     #[#o267, 600.0, "bullet"],
     #[#o270, 600.0, "quotesinglbase"],
     #[#o271, 600.0, "quotedblbase"],
     #[#o272, 600.0, "quotedblright"],
     #[#o273, 600.0, "guillemotright"],
     #[#o274, 600.0, "ellipsis"],
     #[#o275, 600.0, "perthousand"],
     #[#o277, 600.0, "questiondown"],
     #[#o301, 600.0, "grave"],
     #[#o302, 600.0, "acute"],
     #[#o303, 600.0, "circumflex"],
     #[#o304, 600.0, "tilde"],
     #[#o305, 600.0, "macron"],
     #[#o306, 600.0, "breve"],
     #[#o307, 600.0, "dotaccent"],
     #[#o310, 600.0, "dieresis"],
     #[#o312, 600.0, "ring"],
     #[#o313, 600.0, "cedilla"],
     #[#o315, 600.0, "hungarumlaut"],
     #[#o316, 600.0, "ogonek"],
     #[#o317, 600.0, "caron"],
     #[#o320, 600.0, "emdash"],
     #[#o341, 600.0, "AE"],
     #[#o343, 600.0, "ordfeminine"],
     #[#o350, 600.0, "Lslash"],
     #[#o351, 600.0, "Oslash"],
     #[#o352, 600.0, "OE"],
     #[#o353, 600.0, "ordmasculine"],
     #[#o361, 600.0, "ae"],
     #[#o365, 600.0, "dotlessi"],
     #[#o370, 600.0, "lslash"],
     #[#o371, 600.0, "oslash"],
     #[#o372, 600.0, "oe"],
     #[#o373, 600.0, "germandbls"]]);

setup-laserwriter-metrics
  ("Courier-BoldOblique", #"fix", #"bold", #"italic",
   1000, 0.6, #[0, -251, 600, 700],
   #[#[#o040, 600.0, "space"],
     #[#o041, 600.0, "exclam"],
     #[#o042, 600.0, "quotedbl"],
     #[#o043, 600.0, "numbersign"],
     #[#o044, 600.0, "dollar"],
     #[#o045, 600.0, "percent"],
     #[#o046, 600.0, "ampersand"],
     #[#o047, 600.0, "quoteright"],
     #[#o050, 600.0, "parenleft"],
     #[#o051, 600.0, "parenright"],
     #[#o052, 600.0, "asterisk"],
     #[#o053, 600.0, "plus"],
     #[#o054, 600.0, "comma"],
     #[#o055, 600.0, "hyphen"],
     #[#o056, 600.0, "period"],
     #[#o057, 600.0, "slash"],
     #[#o060, 600.0, "zero"],
     #[#o061, 600.0, "one"],
     #[#o062, 600.0, "two"],
     #[#o063, 600.0, "three"],
     #[#o064, 600.0, "four"],
     #[#o065, 600.0, "five"],
     #[#o066, 600.0, "six"],
     #[#o067, 600.0, "seven"],
     #[#o070, 600.0, "eight"],
     #[#o071, 600.0, "nine"],
     #[#o072, 600.0, "colon"],
     #[#o073, 600.0, "semicolon"],
     #[#o074, 600.0, "less"],
     #[#o075, 600.0, "equal"],
     #[#o076, 600.0, "greater"],
     #[#o077, 600.0, "question"],
     #[#o100, 600.0, "at"],
     #[#o101, 600.0, "A"],
     #[#o102, 600.0, "B"],
     #[#o103, 600.0, "C"],
     #[#o104, 600.0, "D"],
     #[#o105, 600.0, "E"],
     #[#o106, 600.0, "F"],
     #[#o107, 600.0, "G"],
     #[#o110, 600.0, "H"],
     #[#o111, 600.0, "I"],
     #[#o112, 600.0, "J"],
     #[#o113, 600.0, "K"],
     #[#o114, 600.0, "L"],
     #[#o115, 600.0, "M"],
     #[#o116, 600.0, "N"],
     #[#o117, 600.0, "O"],
     #[#o120, 600.0, "P"],
     #[#o121, 600.0, "Q"],
     #[#o122, 600.0, "R"],
     #[#o123, 600.0, "S"],
     #[#o124, 600.0, "T"],
     #[#o125, 600.0, "U"],
     #[#o126, 600.0, "V"],
     #[#o127, 600.0, "W"],
     #[#o130, 600.0, "X"],
     #[#o131, 600.0, "Y"],
     #[#o132, 600.0, "Z"],
     #[#o133, 600.0, "bracketleft"],
     #[#o134, 600.0, "backslash"],
     #[#o135, 600.0, "bracketright"],
     #[#o136, 600.0, "asciicircum"],
     #[#o137, 600.0, "underscore"],
     #[#o140, 600.0, "quoteleft"],
     #[#o141, 600.0, "a"],
     #[#o142, 600.0, "b"],
     #[#o143, 600.0, "c"],
     #[#o144, 600.0, "d"],
     #[#o145, 600.0, "e"],
     #[#o146, 600.0, "f"],
     #[#o147, 600.0, "g"],
     #[#o150, 600.0, "h"],
     #[#o151, 600.0, "i"],
     #[#o152, 600.0, "j"],
     #[#o153, 600.0, "k"],
     #[#o154, 600.0, "l"],
     #[#o155, 600.0, "m"],
     #[#o156, 600.0, "n"],
     #[#o157, 600.0, "o"],
     #[#o160, 600.0, "p"],
     #[#o161, 600.0, "q"],
     #[#o162, 600.0, "r"],
     #[#o163, 600.0, "s"],
     #[#o164, 600.0, "t"],
     #[#o165, 600.0, "u"],
     #[#o166, 600.0, "v"],
     #[#o167, 600.0, "w"],
     #[#o170, 600.0, "x"],
     #[#o171, 600.0, "y"],
     #[#o172, 600.0, "z"],
     #[#o173, 600.0, "braceleft"],
     #[#o174, 600.0, "bar"],
     #[#o175, 600.0, "braceright"],
     #[#o176, 600.0, "asciitilde"],
     #[#o241, 600.0, "exclamdown"],
     #[#o242, 600.0, "cent"],
     #[#o243, 600.0, "sterling"],
     #[#o244, 600.0, "fraction"],
     #[#o245, 600.0, "yen"],
     #[#o246, 600.0, "florin"],
     #[#o247, 600.0, "section"],
     #[#o250, 600.0, "currency"],
     #[#o251, 600.0, "quotesingle"],
     #[#o252, 600.0, "quotedblleft"],
     #[#o253, 600.0, "guillemotleft"],
     #[#o254, 600.0, "guilsinglleft"],
     #[#o255, 600.0, "guilsinglright"],
     #[#o256, 600.0, "fi"],
     #[#o257, 600.0, "fl"],
     #[#o261, 600.0, "endash"],
     #[#o262, 600.0, "dagger"],
     #[#o263, 600.0, "daggerdbl"],
     #[#o264, 600.0, "periodcentered"],
     #[#o266, 600.0, "paragraph"],
     #[#o267, 600.0, "bullet"],
     #[#o270, 600.0, "quotesinglbase"],
     #[#o271, 600.0, "quotedblbase"],
     #[#o272, 600.0, "quotedblright"],
     #[#o273, 600.0, "guillemotright"],
     #[#o274, 600.0, "ellipsis"],
     #[#o275, 600.0, "perthousand"],
     #[#o277, 600.0, "questiondown"],
     #[#o301, 600.0, "grave"],
     #[#o302, 600.0, "acute"],
     #[#o303, 600.0, "circumflex"],
     #[#o304, 600.0, "tilde"],
     #[#o305, 600.0, "macron"],
     #[#o306, 600.0, "breve"],
     #[#o307, 600.0, "dotaccent"],
     #[#o310, 600.0, "dieresis"],
     #[#o312, 600.0, "ring"],
     #[#o313, 600.0, "cedilla"],
     #[#o315, 600.0, "hungarumlaut"],
     #[#o316, 600.0, "ogonek"],
     #[#o317, 600.0, "caron"],
     #[#o320, 600.0, "emdash"],
     #[#o341, 600.0, "AE"],
     #[#o343, 600.0, "ordfeminine"],
     #[#o350, 600.0, "Lslash"],
     #[#o351, 600.0, "Oslash"],
     #[#o352, 600.0, "OE"],
     #[#o353, 600.0, "ordmasculine"],
     #[#o361, 600.0, "ae"],
     #[#o365, 600.0, "dotlessi"],
     #[#o370, 600.0, "lslash"],
     #[#o371, 600.0, "oslash"],
     #[#o372, 600.0, "oe"],
     #[#o373, 600.0, "germandbls"]]);


setup-laserwriter-metrics
  ("Symbol", #"symbol", #"normal", #"roman",
   1000, #f, #[-180, -293, 1090, 1010],
   #[#[#o040, 250.0, "space"],
     #[#o041, 333.0, "exclam"],
     #[#o042, 713.0, "universal"],
     #[#o043, 500.0, "numbersign"],
     #[#o044, 549.0, "existential"],
     #[#o045, 833.0, "percent"],
     #[#o046, 778.0, "ampersand"],
     #[#o047, 439.0, "suchthat"],
     #[#o050, 333.0, "parenleft"],
     #[#o051, 333.0, "parenright"],
     #[#o052, 500.0, "asteriskmath"],
     #[#o053, 549.0, "plus"],
     #[#o054, 250.0, "comma"],
     #[#o055, 549.0, "minus"],
     #[#o056, 250.0, "period"],
     #[#o057, 278.0, "slash"],
     #[#o060, 500.0, "zero"],
     #[#o061, 500.0, "one"],
     #[#o062, 500.0, "two"],
     #[#o063, 500.0, "three"],
     #[#o064, 500.0, "four"],
     #[#o065, 500.0, "five"],
     #[#o066, 500.0, "six"],
     #[#o067, 500.0, "seven"],
     #[#o070, 500.0, "eight"],
     #[#o071, 500.0, "nine"],
     #[#o072, 278.0, "colon"],
     #[#o073, 278.0, "semicolon"],
     #[#o074, 549.0, "less"],
     #[#o075, 549.0, "equal"],
     #[#o076, 549.0, "greater"],
     #[#o077, 444.0, "question"],
     #[#o100, 549.0, "congruent"],
     #[#o101, 696.0, "Alpha"],
     #[#o102, 660.0, "Beta"],
     #[#o103, 710.0, "Chi"],
     #[#o104, 612.0, "Delta"],
     #[#o105, 652.0, "Epsilon"],
     #[#o106, 763.0, "Phi"],
     #[#o107, 603.0, "Gamma"],
     #[#o110, 765.0, "Eta"],
     #[#o111, 351.0, "Iota"],
     #[#o112, 631.0, "theta1"],
     #[#o113, 724.0, "Kappa"],
     #[#o114, 686.0, "Lambda"],
     #[#o115, 918.0, "Mu"],
     #[#o116, 739.0, "Nu"],
     #[#o117, 750.0, "Omicron"],
     #[#o120, 768.0, "Pi"],
     #[#o121, 741.0, "Theta"],
     #[#o122, 580.0, "Rho"],
     #[#o123, 592.0, "Sigma"],
     #[#o124, 632.0, "Tau"],
     #[#o125, 690.0, "Upsilon"],
     #[#o126, 439.0, "sigma1"],
     #[#o127, 768.0, "Omega"],
     #[#o130, 645.0, "Xi"],
     #[#o131, 795.0, "Psi"],
     #[#o132, 650.0, "Zeta"],
     #[#o133, 333.0, "bracketleft"],
     #[#o134, 863.0, "therefore"],
     #[#o135, 333.0, "bracketright"],
     #[#o136, 658.0, "perpendicular"],
     #[#o137, 500.0, "underscore"],
     #[#o140, 500.0, "radicalex"],
     #[#o141, 631.0, "alpha"],
     #[#o142, 549.0, "beta"],
     #[#o143, 549.0, "chi"],
     #[#o144, 494.0, "delta"],
     #[#o145, 439.0, "epsilon"],
     #[#o146, 521.0, "phi"],
     #[#o147, 411.0, "gamma"],
     #[#o150, 603.0, "eta"],
     #[#o151, 329.0, "iota"],
     #[#o152, 603.0, "phi1"],
     #[#o153, 549.0, "kappa"],
     #[#o154, 549.0, "lambda"],
     #[#o155, 576.0, "mu"],
     #[#o156, 521.0, "nu"],
     #[#o157, 549.0, "omicron"],
     #[#o160, 549.0, "pi"],
     #[#o161, 521.0, "theta"],
     #[#o162, 549.0, "rho"],
     #[#o163, 603.0, "sigma"],
     #[#o164, 439.0, "tau"],
     #[#o165, 576.0, "upsilon"],
     #[#o166, 713.0, "omega1"],
     #[#o167, 686.0, "omega"],
     #[#o170, 493.0, "xi"],
     #[#o171, 686.0, "psi"],
     #[#o172, 494.0, "zeta"],
     #[#o173, 480.0, "braceleft"],
     #[#o174, 200.0, "bar"],
     #[#o175, 480.0, "braceright"],
     #[#o176, 549.0, "similar"],
     #[#o241, 620.0, "Upsilon1"],
     #[#o242, 247.0, "minute"],
     #[#o243, 549.0, "lessequal"],
     #[#o244, 167.0, "fraction"],
     #[#o245, 713.0, "infinity"],
     #[#o246, 500.0, "florin"],
     #[#o247, 753.0, "club"],
     #[#o250, 753.0, "diamond"],
     #[#o251, 753.0, "heart"],
     #[#o252, 753.0, "spade"],
     #[#o253, 1042.0, "arrowboth"],
     #[#o254, 987.0, "arrowleft"],
     #[#o255, 603.0, "arrowup"],
     #[#o256, 987.0, "arrowright"],
     #[#o257, 603.0, "arrowdown"],
     #[#o260, 400.0, "degree"],
     #[#o261, 549.0, "plusminus"],
     #[#o262, 411.0, "second"],
     #[#o263, 549.0, "greaterequal"],
     #[#o264, 549.0, "multiply"],
     #[#o265, 713.0, "proportional"],
     #[#o266, 494.0, "partialdiff"],
     #[#o267, 460.0, "bullet"],
     #[#o270, 549.0, "divide"],
     #[#o271, 549.0, "notequal"],
     #[#o272, 549.0, "equivalence"],
     #[#o273, 549.0, "approxequal"],
     #[#o274, 1000.0, "ellipsis"],
     #[#o275, 603.0, "arrowvertex"],
     #[#o276, 1000.0, "arrowhorizex"],
     #[#o277, 658.0, "carriagereturn"],
     #[#o300, 823.0, "aleph"],
     #[#o301, 686.0, "Ifraktur"],
     #[#o302, 795.0, "Rfraktur"],
     #[#o303, 987.0, "weierstrass"],
     #[#o304, 768.0, "circlemultiply"],
     #[#o305, 768.0, "circleplus"],
     #[#o306, 823.0, "emptyset"],
     #[#o307, 768.0, "intersection"],
     #[#o310, 768.0, "union"],
     #[#o311, 713.0, "propersuperset"],
     #[#o312, 713.0, "reflexsuperset"],
     #[#o313, 713.0, "notsubset"],
     #[#o314, 713.0, "propersubset"],
     #[#o315, 713.0, "reflexsubset"],
     #[#o316, 713.0, "element"],
     #[#o317, 713.0, "notelement"],
     #[#o320, 768.0, "angle"],
     #[#o321, 713.0, "gradient"],
     #[#o322, 790.0, "registerserif"],
     #[#o323, 790.0, "copyrightserif"],
     #[#o324, 890.0, "trademarkserif"],
     #[#o325, 823.0, "product"],
     #[#o326, 549.0, "radical"],
     #[#o327, 250.0, "dotmath"],
     #[#o330, 713.0, "logicalnot"],
     #[#o331, 603.0, "logicaland"],
     #[#o332, 603.0, "logicalor"],
     #[#o333, 1042.0, "arrowdblboth"],
     #[#o334, 987.0, "arrowdblleft"],
     #[#o335, 603.0, "arrowdblup"],
     #[#o336, 987.0, "arrowdblright"],
     #[#o337, 603.0, "arrowdbldown"],
     #[#o340, 494.0, "lozenge"],
     #[#o341, 329.0, "angleleft"],
     #[#o342, 790.0, "registersans"],
     #[#o343, 790.0, "copyrightsans"],
     #[#o344, 786.0, "trademarksans"],
     #[#o345, 713.0, "summation"],
     #[#o346, 384.0, "parenlefttp"],
     #[#o347, 384.0, "parenleftex"],
     #[#o350, 384.0, "parenleftbt"],
     #[#o351, 384.0, "bracketlefttp"],
     #[#o352, 384.0, "bracketleftex"],
     #[#o353, 384.0, "bracketleftbt"],
     #[#o354, 494.0, "bracelefttp"],
     #[#o355, 494.0, "braceleftmid"],
     #[#o356, 494.0, "braceleftbt"],
     #[#o357, 494.0, "braceex"],
     #[#o360, 790.0, "apple"],
     #[#o361, 329.0, "angleright"],
     #[#o362, 274.0, "integral"],
     #[#o363, 686.0, "integraltp"],
     #[#o364, 686.0, "integralex"],
     #[#o365, 686.0, "integralbt"],
     #[#o366, 384.0, "parenrighttp"],
     #[#o367, 384.0, "parenrightex"],
     #[#o370, 384.0, "parenrightbt"],
     #[#o371, 384.0, "bracketrighttp"],
     #[#o372, 384.0, "bracketrightex"],
     #[#o373, 384.0, "bracketrightbt"],
     #[#o374, 494.0, "bracerighttp"],
     #[#o375, 494.0, "bracerightmid"],
     #[#o376, 494.0, "bracerightbt"]]);


// If you suspect the numbers above, you can run this postscript program
// (probably by copying it to a file and printing that file) and compare
// the numbers that get printed with those of the corresponding fonts and
// characters above.  Remember that the numbers above are scaled to be
// 1000 times as big (second element of the car of the first argument of
// SETUP-LASERWRITER-METRICS is the scale factor) as the widths this
// program will print.

// %!
// % Postscript program to print out the widths of a representative set of 
// % characters in each of several fonts.
// 
// % By Mark Nahabedian, 1/8/90, Symbolics Inc.
// 
// % Sample set of characters toprint out widths for:
// /TheseChars (1AaMN.ilI!) def
// 
// % Sample sets of fonts to print out widths of characters for:
// /TheseFonts [ /Helvetica-BoldOblique /Times-Roman ] def
// /CourierFonts [ /Courier /Courier-Bold /Courier-Oblique /Courier-BoldOblique ] def
// 
// % Text 'cursor' positioning stuff:
// /StartPage { /y 700 def } def
// /NextLine { /y y 20 sub def } def
// /Space { (  ) show } def
// /Tab0 { 30 y moveto } def
// /Tab1 { 178 y moveto } def
// /Tab2 { 230 y moveto } def
// 
// /stringbuf 100 string def
// 
// % single character string interval
// /DoChar {
//	 Tab1 scaled setfont
//	 dup show
//	     Tab2 unscaled setfont
//	     stringwidth pop stringbuf cvs
//	     scaled setfont show NextLine
// } def
// 
// % fontname
// /FontSetup {
//	 findfont dup /unscaled exch def
//		      /scaled exch 10 scalefont def
// } def
// 
// % X Y
// /ShowXandY {
//	 stringbuf cvs show Space
//	 stringbuf cvs show
// } def
// 
// % fontdict
// /ShowFontBBox {
//	 ([ ) show
//	 dup /FontMatrix get dup  3 -1 roll	% stack:  fontdict, FontMatrix, FontMatrix
//	 /FontBBox get 	dup 0 get exch
//			 dup 1 get exch
//	 % stack:  BBox, BBox[1], BBox[0], FontMatrix, FontMatrix
//				 4 1 roll  3 -1 roll
//				 transform exch ShowXandY Space
//			 dup 2 get exch
//			     3 get
//	 % stack:  BBox[3], BBox[2], FontMatrix
//				 3 -1 roll
//				 transform exch ShowXandY Space
//	 ( ]) show
// } def
// 
// % fontname
// /DoFont { 
//	 dup FontSetup
//	     scaled setfont Tab0
//	     stringbuf cvs show Tab1
//	 unscaled ShowFontBBox NextLine
//	 0 1 TheseChars length 1 sub { TheseChars exch 1 getinterval DoChar } for
// } def
// 
// % I tried doing this with forall but it didn't work.
// 
// StartPage
// 0 1 TheseFonts length 1 sub { TheseFonts exch get DoFont } for
// showpage
// 
// StartPage
// 0 1 CourierFonts length 1 sub { CourierFonts exch get DoFont } for
// showpage
